﻿using Simulator.Model;
using Simulator.ViewModels;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.ApplicationModel.Core;
using Windows.Devices.Bluetooth.Rfcomm;
using Windows.Networking.Sockets;
using Windows.Storage.Streams;
using Windows.UI.Core;

namespace Simulator.Bluetooth
{
    class ConnectionManager
    {
        private const ushort runIndex = 10;
        private const ushort runSubindex = 1;
        private RfcommServiceProvider _provider;
        private StreamSocket _socket;
        private object locker = new object();

        IEnumerable<Parameter> parameters;
        IEnumerable<MonitorParameter> monitorParameters;
        Alarm alarm;
        PageViewModel pageViewModel;
        MainPage mainPage;



        public ConnectionManager(IEnumerable<Parameter> parameters, IEnumerable<MonitorParameter> monitorParameters, Alarm alarm, PageViewModel pageViewModel, MainPage mainPage)
        {
            this.parameters = parameters;
            this.monitorParameters = monitorParameters;
            this.alarm = alarm;
            this.pageViewModel = pageViewModel;
            this.mainPage = mainPage;
        }

        public async void Initialize()
        {
            await CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.High, () => {

                this.pageViewModel.State = "Disconnected";
            });
            
            // Initialize the provider for the hosted RFCOMM service
            try
            {
                _provider = await RfcommServiceProvider.CreateAsync(RfcommServiceId.FromUuid(new Guid("3b29c7a5-13b5-42f3-8fb7-8cb80f35aa68")));

                // Create a listener for this service and start listening
                StreamSocketListener listener = new StreamSocketListener();
                listener.ConnectionReceived += OnConnectionReceived;

                var rfcomm = _provider.ServiceId.AsString();

                await listener.BindServiceNameAsync(
                    _provider.ServiceId.AsString(),
                    SocketProtectionLevel
                        .BluetoothEncryptionAllowNullAuthentication);

                // Set the SDP attributes and start advertising
                InitializeServiceSdpAttributes(_provider);
                _provider.StartAdvertising(listener, true);
            }
            catch (Exception e)
            {
                Debug.Write(e.Message);
            }
        }

        const uint SERVICE_VERSION_ATTRIBUTE_ID = 0x0300;
        const byte SERVICE_VERSION_ATTRIBUTE_TYPE = 0x0A;   // UINT32
        const uint SERVICE_VERSION = 200;

        const UInt16 SdpServiceNameAttributeId = 0x100;
        const byte SdpServiceNameAttributeType = (4 << 3) | 5;
        const string SdpServiceName = "Bluetooth Rfcomm Inverter Service";

        void InitializeServiceSdpAttributes(RfcommServiceProvider provider)
        {
            DataWriter writer = new DataWriter();

            // First write the attribute type
            writer.WriteByte(SERVICE_VERSION_ATTRIBUTE_TYPE);
            // Then write the data
            writer.WriteUInt32(SERVICE_VERSION);

            IBuffer data = writer.DetachBuffer();
            provider.SdpRawAttributes.Add(SERVICE_VERSION_ATTRIBUTE_ID, data);
        }

        async void OnConnectionReceived(
            StreamSocketListener listener,
            StreamSocketListenerConnectionReceivedEventArgs args)
        {
            // Stop advertising/listening so that we're only serving one client
            int value = 12;
            Random rnd = new Random();
            bool connected = true;
            _provider.StopAdvertising();

            try
            {
                _socket = args.Socket;
            }
            catch (Exception e)
            {
                Debug.Write(e);
                return;
            }

            await CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.High, () => {
                // Asign data context once it is connected
                this.mainPage.DataContext = this.pageViewModel;
                this.pageViewModel.State = "Connected";
            });

            // The client socket is connected. At this point the App can wait for
            // the user to take some action, e.g. click a button to receive a file
            // from the device, which could invoke the Picker and then save the
            // received file to the picked location. The transfer itself would use
            // the Sockets API and not the Rfcomm API, and so is omitted here for
            // brevity.


            var writer = new DataWriter(_socket.OutputStream);
            var reader = new DataReader(_socket.InputStream);
            reader.InputStreamOptions = InputStreamOptions.Partial;
            while (connected)
            {
                try
                {
                    // Wait listening
                    uint typeSize = await reader.LoadAsync(1);
                    var typeBytes = new byte[typeSize];
                    reader.ReadBytes(typeBytes);

                    byte type = typeBytes[0];

                    uint indexSize = await reader.LoadAsync(sizeof(ushort));
                    var indexBytes = new byte[indexSize];
                    reader.ReadBytes(indexBytes);
                    ushort index = BitConverter.ToUInt16(indexBytes, 0);

                    uint subindexSize = await reader.LoadAsync(sizeof(ushort));
                    var subindexBytes = new byte[subindexSize];
                    reader.ReadBytes(subindexBytes);
                    ushort subindex = BitConverter.ToUInt16(subindexBytes, 0);

                    uint dataSize = await reader.LoadAsync(sizeof(int));
                    var dataBytes = new byte[dataSize];
                    reader.ReadBytes(dataBytes);
                    int data = BitConverter.ToInt32(dataBytes, 0);

                    List<byte> response = new List<byte>();
                    byte[] valueArray;
                    // Depending what kind of action, different stuff to do
                    switch (type)
                    {

                        case 2:
                            Parameter parameter = parameters.SingleOrDefault(x => x.Index == index && x.Subindex == subindex);
                            if(parameter != null)
                            {
                                response.Add(0);

                                valueArray = BitConverter.GetBytes(parameter.Value.Value);
                                response.Add((byte)valueArray.Length);
                                response.AddRange(valueArray);
                                break;
                            }

                            MonitorParameter monitorParameter = monitorParameters.SingleOrDefault(x => x.Index == index && x.Subindex == subindex);
                            if (monitorParameter != null)
                            {
                                response.Add(0);
                                valueArray = BitConverter.GetBytes(monitorParameter.Value.Value);
                                response.Add((byte)valueArray.Length);
                                response.AddRange(valueArray);
                                break;
                            }

                            if (index == 1 && subindex == 1)
                            {
                                response.Add(0);
                                valueArray = Encoding.UTF8.GetBytes("Inverter1_1_1");
                                response.Add((byte)valueArray.Length);
                                response.AddRange(valueArray);
                                break;
                            }

                            if (alarm.Index == index && alarm.SubIndex == subindex)
                            {
                                response.Add(0);
                                valueArray = BitConverter.GetBytes(alarm.Value);
                                response.Add((byte)valueArray.Length);
                                response.AddRange(valueArray);
                                break;
                            }

                            if(index == 9000)
                            {
                                if(subindex == 1)
                                {
                                    response.Add(0);
                                    valueArray = Encoding.UTF8.GetBytes(pageViewModel.Model);
                                    response.Add((byte)valueArray.Length);
                                    response.AddRange(valueArray);
                                }
                                else if (subindex == 2)
                                {
                                    response.Add(0);
                                    valueArray = Encoding.UTF8.GetBytes(pageViewModel.Firmware);
                                    response.Add((byte)valueArray.Length);
                                    response.AddRange(valueArray);
                                }
                                else if (subindex == 3)
                                {
                                    response.Add(0);
                                    valueArray = Encoding.UTF8.GetBytes(pageViewModel.SerialNumber);
                                    response.Add((byte)valueArray.Length);
                                    response.AddRange(valueArray);
                                }
                                else if (subindex == 4)
                                {
                                    response.Add(0);
                                    valueArray = Encoding.UTF8.GetBytes(pageViewModel.Motor);
                                    response.Add((byte)valueArray.Length);
                                    response.AddRange(valueArray);
                                }
                                break;
                            }

                            // Error
                            response.Add(1);
                            valueArray = BitConverter.GetBytes(0);
                            response.Add((byte)valueArray.Length);
                            response.AddRange(valueArray);

                            break;
                        case 3:
                            Parameter parameter1 = parameters.SingleOrDefault(x => x.Index == index && x.Subindex == subindex);
                            if (parameter1 != null)
                            {
                                response.Add(0);
                                valueArray = BitConverter.GetBytes(0);
                                response.Add((byte)valueArray.Length);
                                response.AddRange(valueArray);
                                await CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.High, () => {
                                
                                    parameter1.Value = data;
                                });
                                
                                break;
                            }

                            if (alarm.Index == index && alarm.SubIndex == subindex)
                            {
                                response.Add(0);
                                valueArray = BitConverter.GetBytes(0);
                                response.Add((byte)valueArray.Length);
                                response.AddRange(valueArray);
                                await CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.High, () => {

                                    this.pageViewModel.SelectedAlarm = 0;
                                });
                                
                                break;
                            }

                            if (runIndex == index && runSubindex == subindex)
                            {
                                response.Add(0);
                                valueArray = BitConverter.GetBytes(0);
                                response.Add((byte)valueArray.Length);
                                response.AddRange(valueArray);
                                await CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.High, () => {
                                    if(data == 0)
                                    {
                                        this.pageViewModel.IsRunning = false;
                                        this.pageViewModel.IsStopped = true;
                                    }
                                    else
                                    {
                                        this.pageViewModel.IsRunning = true;
                                        this.pageViewModel.IsStopped = false;
                                    }

                                    
                                });

                                break;
                            }

                            // Error
                            response.Add(1);
                            valueArray = BitConverter.GetBytes(0);
                            response.Add((byte)valueArray.Length);
                            response.AddRange(valueArray);

                            break;

                        case 4:
                            response.Add(0);
                            valueArray = BitConverter.GetBytes((int)rnd.Next(-100, 235));
                            response.Add((byte)valueArray.Length);
                            response.AddRange(valueArray);
                            value = data;
                            break;

                        case 5:
                            response.Add(0);
                            valueArray = BitConverter.GetBytes(0);
                            response.Add((byte)valueArray.Length);
                            response.AddRange(valueArray);

                            await CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.High, () => {
                                foreach (Parameter item in parameters)
                                {
                                    item.Value = item.Default;
                                }});

                            break;
                    }

                    
                    writer.WriteBytes(response.ToArray());
                    await writer.StoreAsync();
                }
                catch (Exception e)
                {
                    Debug.Write(e.Message);
                    connected = false;
                }
            }

            this._provider = null;
            this.Initialize();
        }
    }


}